/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <stdio.h>

/*****************************************************************************
 * Qt Headers
 *****************************************************************************/

#include <qapplication.h>
#include <qfontmetrics.h>
#include <qpaintdevicemetrics.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qpainter.h>
#include <qbrush.h>
#include <qpen.h>
#include <qcolor.h>
#include <qrect.h>
#include <qpixmap.h>
#include <qtimer.h>
#include <qprinter.h>
#include <qmessagebox.h>
#include <qstack.h>
#include <qmap.h>

/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include "qscale.h"
#include "qtag.h"
#include "qtagvisitor.h"
#include "qscalereader.h"

/*****************************************************************************
 * Local defines, enums, constants and typedefs
 *****************************************************************************/

#define RND(x) (int)((x) >= 0.0 ? (x) + 0.5 : (x) - 0.5)

typedef enum
{
   A, A_SHARP, B, C, C_SHARP, D, D_SHARP, E, F, F_SHARP, G, G_SHARP

} Notes;

const char *keys[] =
{
   "A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"
};

const char *fkeys[] =
{
   "A", "Bb", "B", "C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab"
};

static const int  NOTES      = sizeof(keys) / sizeof(keys[0]);
static const int  NSTRINGS   = 6;
static const char *FILE_NAME = "qscale.xml";

typedef QStack<QPopupMenu> MenuStack;
typedef QMap<int, const QTag *>  EntryMap;

/*****************************************************************************
 * Local class: MenuVisitor
 *****************************************************************************/

class MenuVisitor : public QTagVisitor
{
   public:
      MenuVisitor(QScale *theObject, MenuStack *stack);
      ~MenuVisitor();
      void begin(const QTag *theTag);
      void end(const QTag *theTag);
      EntryMap *take();

   private:
      MenuVisitor(const MenuVisitor &);
      MenuVisitor &operator=(const MenuVisitor &);

      int isFolderTag(const QTag *theTag);

   private:
      EntryMap *mMap;
      QScale *mObject;
      MenuStack *mStack;
      int mIndex;
};

MenuVisitor::MenuVisitor(QScale *theObject, MenuStack *stack) :
   mMap(new EntryMap()),
   mObject(theObject),
   mStack(stack),
   mIndex(0)
{
}

MenuVisitor::~MenuVisitor()
{
   delete mMap;
}

void MenuVisitor::begin(const QTag *theTag)
{
   // Ignore the main tag

   if(theTag->name() != QSCALE)
   {
      if(isFolderTag(theTag))
      {
         QPopupMenu *menu = new QPopupMenu(mStack->top());

         mStack->top()->insertItem(theTag->getAttribute(NAME), menu);
         mStack->push(menu);

         QObject::connect(
            menu,
            SIGNAL(activated(int)),
            mObject,
            SLOT(slotScale(int))
         );
      }
      else
      {
         // It's a scale entry

         mStack->top()->insertItem(
            theTag->getAttribute(SCALE),
            mIndex
         );

         mMap->insert(mIndex++, theTag);
      }
   }
}

void MenuVisitor::end(const QTag *theTag)
{
   if(isFolderTag(theTag))
      mStack->pop();
}

EntryMap *MenuVisitor::take()
{
   EntryMap *map = mMap;
   mMap          = NULL;

   return(map);
}

int MenuVisitor::isFolderTag(const QTag *theTag)
{
   return(!theTag->getAttribute(NAME).isNull());
}

/*****************************************************************************
 * Local Struct: QScale::Internal
 *****************************************************************************/

struct QScale::Internal
{
   Internal(QTag *theTag, EntryMap *map);
   ~Internal();

   void setCurrentScale(int index);

   QPopupMenu *mKeys;
   QTag       *mTag;
   int        mCurrentKey;
   int        mCurrentScale[NOTES];
   int        mCurrentScaleIndex;
   QWidget    *mWidget;
   QString    mCurrentScaleName;
   QTimer     *mTimer;
   EntryMap   *mMap;
   bool       mUseSharps;
   int        mUseMenuID;
   QPopupMenu *mSelect;
   int        mKeysFirstID;
};

QScale::Internal::Internal(QTag *theTag, EntryMap *map) :
   mKeys(0),
   mTag(theTag),
   mCurrentKey(0),
   mCurrentScaleIndex(0),
   mWidget(0),
   mCurrentScaleName(""),
   mTimer(0),
   mMap(map),
   mUseSharps(true),
   mUseMenuID(0),
   mSelect(0),
   mKeysFirstID(0)
{
   setCurrentScale(0);
}

QScale::Internal::~Internal()
{
   delete mTag;
   delete mTimer;
   delete mMap;
}

void QScale::Internal::setCurrentScale(int index)
{
   QString values = (*mMap)[index]->getAttribute(PATTERN);

   for(int i = 0; i < NOTES; i ++)
   {
      if(values.at(i) == ' ')
         mCurrentScale[i] = 0;
      else
         mCurrentScale[i] = 1;
   }

   mCurrentScaleName  = (*mMap)[index]->getAttribute(SCALE);
   mCurrentScaleIndex = index;
}

/*****************************************************************************
 * Local Class: DrawingWidget
 *****************************************************************************/

class DrawingWidget : public QWidget
{
   public:
      DrawingWidget(QWidget *parent, QScale *scale);

   protected:
      void paintEvent(QPaintEvent *e);
      void resizeEvent(QResizeEvent *e);

   private:
      QScale *mScale;
};

DrawingWidget::DrawingWidget(QWidget *parent, QScale *scale) :
   QWidget(parent),
   mScale(scale)
{
}

void DrawingWidget::paintEvent(QPaintEvent *e)
{
   mScale->handlePaintEvent(this, e);
}

void DrawingWidget::resizeEvent(QResizeEvent *e)
{
   mScale->handleResizeEvent(this, e);
}

/*****************************************************************************
 * Definition of Class: QScale
 *****************************************************************************/

QScale::QScale(
   QTag *theTag,
   QWidget *parent,
   const char *name,
   WFlags f
) :
   QMainWindow(parent, name, f),
   mInternal(0)
{
   resize(850, 500);

   // Create the 'File' menu ...

   QPopupMenu *file = new QPopupMenu();
   menuBar()->insertItem("&File", file);

   file->insertItem(
      "Print",
      this,
      SLOT(slotPrint()),
      CTRL + Key_P
   );

   file->insertItem(
      "Quit",
      this,
      SLOT(slotQuit()),
      CTRL + Key_Q
   );

   // Ok, now the 'Select' menu ...

   QPopupMenu *select = new QPopupMenu();
   select->insertTearOffHandle();

   menuBar()->insertItem("&Select", select);

   int useMenuID = select->insertItem(
      "Use Flats",
      this,
      SLOT(slotSharpsFlats()),
      CTRL + Key_X
   );

   select->insertItem(
      "Key >>",
      this,
      SLOT(slotKeyIncrement()),
      CTRL + Key_Right
   );

   select->insertItem(
      "Key <<",
      this,
      SLOT(slotKeyDecrement()),
      CTRL + Key_Left
   );

   QPopupMenu *keysMenu = new QPopupMenu(select);

   select->insertItem("&Key", keysMenu);

   QObject::connect(
      keysMenu,
      SIGNAL(activated(int)),
      this,
      SLOT(slotKey(int))
   );

   int i           = 0;
   int keysFirstID = -1;

   for(i = 0; i < NOTES; i ++)
   {
      int theID = keysMenu->insertItem(keys[i], i);

      if(keysFirstID == -1)
         keysFirstID = theID;
   }

   select->insertItem(
      "Scale >>",
      this,
      SLOT(slotScaleIncrement()),
      ALT + Key_Right
   );

   select->insertItem(
      "Scale <<",
      this,
      SLOT(slotScaleDecrement()),
      ALT + Key_Left
   );

   QPopupMenu *scales = new QPopupMenu(select);
   select->insertItem("&Scale", scales);

   QObject::connect(
      scales,
      SIGNAL(activated(int)),
      this,
      SLOT(slotScale(int))
   );

   MenuStack stack;
   stack.push(scales);

   MenuVisitor visitor(this, &stack);
   theTag->accept(visitor);

   mInternal = new Internal(
      theTag,
      visitor.take()
   );

   mInternal->mKeys  = keysMenu;
   mInternal->mTimer = new QTimer(this);

   QObject::connect(
      mInternal->mTimer,
      SIGNAL(timeout()),
      this,
      SLOT(slotTimerFinished())
   );

   mInternal->mWidget = new DrawingWidget(this, this);
   mInternal->mSelect = select;

   setCentralWidget(mInternal->mWidget);

   mInternal->mUseMenuID   = useMenuID;
   mInternal->mKeysFirstID = keysFirstID;
}

QScale::~QScale()
{
   delete mInternal;
}

void QScale::handlePaintEvent(QWidget *target, QPaintEvent *)
{
   if(mInternal->mTimer->isActive())
      return;

   int width  = target->width();
   int height = target->height();

   QPixmap pixmap(width, height);
   QPainter p(&pixmap);

   drawFretBoard(&p, width, height);

   QPainter screen(target);
   screen.drawPixmap(0, 0, pixmap);
}

void QScale::handleResizeEvent(QWidget *, QResizeEvent *)
{
   if(mInternal->mTimer->isActive())
      mInternal->mTimer->stop();

   mInternal->mTimer->start(250, TRUE);
}

void QScale::slotQuit()
{
   qApp->quit();
}

void QScale::slotKey(int which)
{
   mInternal->mCurrentKey = which;
   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotScale(int which)
{
   mInternal->setCurrentScale(which);
   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotTimerFinished()
{
   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotPrint()
{
   QPrinter thePrinter;

   if(thePrinter.setup(this))
   {
      QPainter p(&thePrinter);
      QPaintDeviceMetrics metrics(&thePrinter);

      drawFretBoard(&p, metrics.width(), metrics.height());
   }
}

void QScale::slotKeyIncrement()
{
   if(++mInternal->mCurrentKey == NOTES)
      mInternal->mCurrentKey = 0;

   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotKeyDecrement()
{
   if(--mInternal->mCurrentKey == -1)
      mInternal->mCurrentKey = NOTES - 1;

   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotScaleIncrement()
{
   if(++mInternal->mCurrentScaleIndex == (int)mInternal->mMap->count())
      mInternal->mCurrentScaleIndex = 0;

   mInternal->setCurrentScale(mInternal->mCurrentScaleIndex);
   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotScaleDecrement()
{
   if(--mInternal->mCurrentScaleIndex == -1)
      mInternal->mCurrentScaleIndex = mInternal->mMap->count() - 1;

   mInternal->setCurrentScale(mInternal->mCurrentScaleIndex);
   mInternal->mWidget->repaint(FALSE);
}

void QScale::slotSharpsFlats()
{
   if(mInternal->mUseSharps)
   {
      // Switch to flats ...

      mInternal->mSelect->changeItem(mInternal->mUseMenuID, "Use Sharps");
      mInternal->mUseSharps = false;

      for(int i = 0; i < NOTES; i ++)
      {
         mInternal->mKeys->changeItem(
            mInternal->mKeysFirstID + i,
            fkeys[i]
         );
      }
   }
   else
   {
      // Switch to sharps ...

      mInternal->mSelect->changeItem(mInternal->mUseMenuID, "Use Flats");
      mInternal->mUseSharps = true;

      for(int i = 0; i < NOTES; i ++)
      {
         mInternal->mKeys->changeItem(
            mInternal->mKeysFirstID + i,
            keys[i]
         );
      }
   }

   mInternal->mWidget->repaint(FALSE);
}

void QScale::drawFretBoard(QPainter *p, int width, int height)
{
   QBrush brush(QColor(255, 255, 255));
   QRect rect(0, 0, width, height);

   p->fillRect(rect, brush);

   double w  = .9 * width;
   double h  = .2 * w;
   double x0 = (width - w)  / 2.0;
   double y0 = (height - h) / 2.0;

   int i        = 0;
   double y     = y0;
   double yStep = h / 5.0;

   for(i = 0; i < 6; i ++)
   {
      p->drawLine(RND(x0), RND(y), RND(x0 + w), RND(y));
      y += yStep;
   }

   double x     = x0;
   double xStep = w / 24.0;

   for(i = 0; i <= 24; i ++)
   {
      if(i % NOTES == NOTES - 1 || i % NOTES == 0)
         p->setPen(QPen(QColor(0, 0, 0), 3));
      else
         p->setPen(QPen(QColor(0, 0, 0), 1));

      p->drawLine(RND(x), RND(y0), RND(x), RND(y0 + h));
      x += xStep;
   }

   y     = y0;
   yStep = h / 5.0;

   double radius = yStep / 1.3;

   p->setPen(QPen(QColor(0, 0, 0), 1));

   static int strings[NSTRINGS] =
   {
      E, A, D, G, B, E
   };

   double factor1 = 0.5;
   double factor2 = 0.85;
   double divisor = 2.0;

#ifdef _WIN32
   divisor *= 2.0;
#endif

   QFont theFont("courier");
   theFont.setPointSizeFloat(factor1 * radius);
   p->setFont(theFont);

   QFontMetrics metrics = p->fontMetrics();
   const char **theKeys = (mInternal->mUseSharps ? keys : fkeys);

   for(i = 0; i < 6; i ++)
   {
      xStep = w / 24.0;
      x     = x0 - xStep / 2.0;

      for(int j = 0; j <= 24; j ++)
      {
         // Is this note in the scale?

         int note  = (strings[5 - i] + j) % NOTES;
         int index = (note + (NOTES - mInternal->mCurrentKey)) % NOTES;

         if(mInternal->mCurrentScale[index])
         {
            p->setPen(QPen(QColor(0, 0, 0), QPen::NoPen));
            p->setBrush(QBrush(QColor(255, 255, 255), QBrush::SolidPattern));

            p->drawEllipse(
               RND(x - radius / 2.0),
               RND(y - radius / 2.0),
               RND(radius),
               RND(radius)
            );

            p->setPen(QPen(QColor(0, 0, 0), 1));
            p->setBrush(QBrush(QBrush::NoBrush));

            p->drawEllipse(
               RND(x - radius / 2.0),
               RND(y - radius / 2.0),
               RND(radius),
               RND(radius)
            );

            const char *text = theKeys[note];

            QRect bound = metrics.boundingRect(text);

            p->drawText(
               RND(x - bound.width() / 2.0),
               RND(y + bound.height() / divisor),
               text
            );
         }

         x += xStep;
      }

      y += yStep;
   }

   QString caption;

   caption.sprintf(
      "Key of: %s %s [",
      theKeys[mInternal->mCurrentKey],
      mInternal->mCurrentScaleName.latin1()
   );

   int first = 1;

   for(i = 0; i < NOTES; i ++)
   {
      if(mInternal->mCurrentScale[i])
      {
         if(first)
            first = 0;
         else
            caption.append(" ");

         caption.append(theKeys[(i + mInternal->mCurrentKey) % NOTES]);
      }
   }

   caption.append("]");

   theFont.setPointSizeFloat(factor2 * radius);
   p->setFont(theFont);
   metrics = p->fontMetrics();

   QRect bound = metrics.boundingRect(caption);

   p->drawText(
      RND(x0 + (w - bound.width()) / 2.0),
      RND((y0 - bound.height()) / 2.0),
      caption
   );

   p->end();
}

int main(int argc, char **argv)
{
   QApplication theApplication(argc, argv);

   QString errMsg("");

   QTag *theTag = QScaleReader::read(
      argc > 1 ? argv[1] : FILE_NAME,
      NOTES,
      errMsg
   );

   if(theTag == NULL)
   {
      QMessageBox::message("Error", errMsg);
      theTag = QScaleReader::getDefault();
   }

   QScale theObject(theTag);

   theApplication.setMainWidget(&theObject);
   theObject.show();

   return(theApplication.exec());
}
